@@ -1,9 +1,10 @@
+from __future__ import absolute_import, division, print_function
 import urwid
-
-
+from urwid.util import _target_encoding, calc_width
 
 
 # generic urwid helpers -------------------------------------------------------
+
 def make_canvas(txt, attr, maxcol, fill_attr=None):
     processed_txt = []
     processed_attr = []
@@ -13,7 +14,7 @@ def make_canvas(txt, attr, maxcol, fill_attr=None):
         # filter out zero-length attrs
         line_attr = [(aname, l) for aname, l in line_attr if l > 0]
 
-        diff = maxcol - len(line)
+        diff = maxcol - calc_width(line, 0, len(line))
         if diff > 0:
             line += " "*diff
             line_attr.append((fill_attr, diff))
@@ -23,9 +24,20 @@ def make_canvas(txt, attr, maxcol, fill_attr=None):
             line_attr = rle_subseg(line_attr, 0, maxcol)
 
         from urwid.util import apply_target_encoding
-        line, line_cs = apply_target_encoding(line)
+        encoded_line, line_cs = apply_target_encoding(line)
 
-        processed_txt.append(line)
+        # line_cs contains byte counts as requested by TextCanvas, but
+        # line_attr still contains column counts at this point: let's fix this.
+        def get_byte_line_attr(line, line_attr):
+            i = 0
+            for label, column_count in line_attr:
+                byte_count = len(line[i:i+column_count].encode(_target_encoding))
+                i += column_count
+                yield label, byte_count
+
+        line_attr = list(get_byte_line_attr(line, line_attr))
+
+        processed_txt.append(encoded_line)
         processed_attr.append(line_attr)
         processed_cs.append(line_cs)
 
@@ -36,8 +48,6 @@ def make_canvas(txt, attr, maxcol, fill_attr=None):
             maxcol=maxcol)
 
 
-
-
 def make_hotkey_markup(s):
     import re
     match = re.match(r"^([^_]*)_(.)(.*)$", s)
@@ -50,17 +60,12 @@ def make_hotkey_markup(s):
             ]
 
 
-
-
-
 def labelled_value(label, value):
     return urwid.AttrMap(urwid.Text([
         ("label", label), str(value)]),
         "fixed value", "fixed value")
 
 
-
-
 class SelectableText(urwid.Text):
     def selectable(self):
         return True
@@ -69,19 +74,28 @@ class SelectableText(urwid.Text):
         return key
 
 
-
 class SignalWrap(urwid.WidgetWrap):
-    def __init__(self, w):
+    def __init__(self, w, is_preemptive=False):
         urwid.WidgetWrap.__init__(self, w)
         self.event_listeners = []
+        self.is_preemptive = is_preemptive
 
     def listen(self, mask, handler):
         self.event_listeners.append((mask, handler))
 
     def keypress(self, size, key):
-        result = self._w.keypress(size, key)
+        result = key
+
+        if self.is_preemptive:
+            for mask, handler in self.event_listeners:
+                if mask is None or mask == key:
+                    result = handler(self, size, key)
+                    break
 
         if result is not None:
+            result = self._w.keypress(size, key)
+
+        if result is not None and not self.is_preemptive:
             for mask, handler in self.event_listeners:
                 if mask is None or mask == key:
                     return handler(self, size, key)
@@ -89,9 +103,8 @@ class SignalWrap(urwid.WidgetWrap):
         return result
 
 
+# {{{ debugger-specific stuff
 
-
-# debugger-specific stuff -----------------------------------------------------
 class StackFrame(urwid.FlowWidget):
     def __init__(self, is_current, name, class_name, filename, line):
         self.is_current = is_current
@@ -135,11 +148,15 @@ class StackFrame(urwid.FlowWidget):
     def keypress(self, size, key):
         return key
 
+
 class BreakpointFrame(urwid.FlowWidget):
-    def __init__(self, is_current, filename, line):
+    def __init__(self, is_current, filename, breakpoint):
         self.is_current = is_current
         self.filename = filename
-        self.line = line
+        self.breakpoint = breakpoint
+        self.line = breakpoint.line  # Starts at 1
+        self.enabled = breakpoint.enabled
+        self.hits = breakpoint.hits
 
     def selectable(self):
         return True
@@ -154,14 +171,18 @@ class BreakpointFrame(urwid.FlowWidget):
         else:
             apfx = ""
 
+        bp_pfx = ''
+        if not self.enabled:
+            apfx += "disabled "
+            bp_pfx += "X"
         if self.is_current:
             apfx += "current "
-            crnt_pfx = ">> "
-        else:
-            crnt_pfx = "   "
+            bp_pfx += ">>"
+        bp_pfx = bp_pfx.ljust(3)
 
-        loc = " %s:%d" % (self.filename, self.line)
-        text = crnt_pfx+loc
+        hits_label = 'hits' if self.hits != 1 else 'hit'
+        loc = " %s:%d (%s %s)" % (self.filename, self.line, self.hits, hits_label)
+        text = bp_pfx+loc
         attr = [(apfx+"breakpoint", len(loc))]
 
         return make_canvas([text], [attr], maxcol, apfx+"breakpoint")
@@ -170,8 +191,6 @@ class BreakpointFrame(urwid.FlowWidget):
         return key
 
 
-
-
 class SearchController(object):
     def __init__(self, ui):
         self.ui = ui
@@ -185,7 +204,6 @@ class SearchController(object):
             self.highlight_line.set_highlight(False)
             self.highlight_line = None
 
-
     def cancel_search(self):
         self.cancel_highlight()
         self.hide_search_ui()
@@ -207,7 +225,7 @@ class SearchController(object):
 
             lhs_col.item_types.insert(
                     0, ("flow", None))
-            lhs_col.widget_list.insert( 0, self.search_AttrMap)
+            lhs_col.widget_list.insert(0, self.search_AttrMap)
 
             self.ui.columns.set_focus(lhs_col)
             lhs_col.set_focus(self.search_AttrMap)
@@ -264,11 +282,9 @@ class SearchController(object):
         return False
 
 
-
-
 class SearchBox(urwid.Edit):
     def __init__(self, controller):
-        urwid.Edit.__init__(self, [("label", "Search: ") ], "")
+        urwid.Edit.__init__(self, [("label", "Search: ")], "")
         self.controller = controller
 
     def restart_search(self):
@@ -300,7 +316,9 @@ class SearchBox(urwid.Edit):
             if self.controller.perform_search(dir=1, s=txt):
                 self.controller.search_AttrMap.set_attr_map({None: "search box"})
             else:
-                self.controller.search_AttrMap.set_attr_map({None: "search not found"})
+                self.controller.search_AttrMap.set_attr_map(
+                        {None: "search not found"})
 
         return result
 
+# }}}
